
#include <stdint.h>

#include <stdbool.h>

#include <gba_input.h>
#include <gba_interrupt.h>
#include <gba_systemcalls.h>

#include "gbt_player/gbt_player.h"

#include "pmd_ad.h"

extern const uint8_t *template[];

extern const unsigned char *squirtledata[];
extern const unsigned char *pmd_titleBitmap[];

#define REG_DISPLAYCONTROL   *((volatile uint32_t *)0x04000000)
#define VIDEOMODE_3          0x0003
#define BGMODE_2             0x0400

#define MEM_VRAM             ((volatile uint16_t *)0x06000000)
#define MEM_VRAMOBJ          ((volatile uint16_t *)0x06014000)
#define MEM_OAM              ((volatile uint16_t *)0x07000000)
#define MEM_CRAM             ((volatile uint16_t *)0x05000000)
#define MEM_CRAMOBJ          ((volatile uint16_t *)0x05000200)
#define SCREEN_WIDTH         240
#define SCREEN_HEIGHT        160

#define KEY_A                1
#define KEY_B                (1 << 1)
#define KEY_SELECT           (1 << 2)
#define KEY_START            (1 << 3)
#define KEY_RIGHT            (1 << 4)
#define KEY_LEFT             (1 << 5)
#define KEY_UP               (1 << 6)
#define KEY_DOWN             (1 << 7)

typedef unsigned char  byte;
typedef unsigned short word;

const word targetPos[] = {
40,60,120,20,180,80,100,60,224,16,16,16,175,100,190,30
};

// UnCompress LZ77 Data to WRAM
inline void LZ77UnCompWRAM(u32 source, u32 dest) {
asm("mov r0, %0\n"
"mov r1, %1\n"
"swi 0x11\n"
:
:"r" (source), "r" (dest)
:"r0", "r1" );
}


// UnCompress LZ77 Data to VRAM
inline void LZ77UnCompVRAM(u32 source, u32 dest) {
asm("mov r0, %0\n"
"mov r1, %1\n"
"swi 0x12\n"
:
:"r" (source), "r" (dest)
:"r0", "r1" );
}

// Using BGR15
uint16_t make_color(unsigned r, unsigned g, unsigned b) {
    return r | (g << 5) | (b << 10);
}

void drawRectangle(int left, int top, int right, int bottom, uint16_t color){
    word top_offset,bottom_offset,i,temp;

    if(top>bottom){
        temp=top;
        top=bottom;
        bottom=temp;
    }
    if(left>right){
        temp=left;
        left=right;
        right=temp;
    }

    top_offset=(top<<8)-(top<<4);
    bottom_offset=(bottom<<8)-(bottom<<4);

    for(i=left;i<=right;i++){
        MEM_VRAM[top_offset+i]=color;
        MEM_VRAM[bottom_offset+i]=color;
    }
    for(i=top_offset;i<=bottom_offset;i+=240){
        MEM_VRAM[left+i]=color;
        MEM_VRAM[right+i]=color;
    }
}

void fillRectangle(int left, int top, int right, int bottom, uint16_t color){
    word top_offset,bottom_offset,i,temp,i2;

    if(top>bottom){
        temp=top;
        top=bottom;
        bottom=temp;
    }
    if(left>right){
        temp=left;
        left=right;
        right=temp;
    }

    top_offset=(top<<8)-(top<<4);
    bottom_offset=(bottom<<8)-(bottom<<4);

    for(i=top_offset;i<=bottom_offset;i+=240){
        for(i2=i+left;i2<=i+right;i2++){
                MEM_VRAM[i2]=color;
        }
    }
}

//stolen from https://github.com/GRRLIB/GRRLIB/blob/v4.5.1/GRRLIB/GRRLIB/grrlib/GRRLIB_collision.h
bool check_point_collision (x1, y1, w1, h1, x2, y2)
     int x1, y1, w1, h1, x2, y2;
{
  return( ((x2>=x1) && (x2<=(x1+w1))) &&
            ((y2>=y1) && (y2<=(y1+h1)))  );
}

// Entry point
int main(void) {
    int i;
    uint16_t oldinput;
    uint16_t keyinput;
    uint16_t newinput;
    uint16_t backupCo;
    int x, y;
    int tx, ty;
    byte tt = 0;
    byte ttn = 0;
    byte score = 0;
    byte tTimer;
    byte tLoc;
    byte gameTimerFrames = 0;
    word gameTimer = 0;
    unsigned char watergunTimer = 0;

    irqInit();
    irqSet(IRQ_VBLANK, gbt_update);
    irqEnable(IRQ_VBLANK);

    // Start at center
    x = SCREEN_WIDTH / 2;
    y = SCREEN_HEIGHT / 2;
    tx = SCREEN_WIDTH / 2;
    ty = SCREEN_HEIGHT / 2;
    tTimer = 200;
    tLoc = 0;

    // Activate video mode 3, to allow direct access to 16-bits pixel buffer
    VBlankIntrWait();
    REG_DISPLAYCONTROL = VIDEOMODE_3 | BGMODE_2 | 0x1000 | (1 << 7);

    LZ77UnCompVRAM((u32)squirtledata, (u32)MEM_VRAMOBJ);

    // Paint screen black
    for (i = 0; i < SCREEN_WIDTH * SCREEN_HEIGHT; ++i)
        MEM_VRAM[i] = make_color(0, 0, 0);

    for (i = 0; i < 0x400; ++i)
        MEM_OAM[i] = 0x0200;

    MEM_OAM[3] = 128;
    MEM_OAM[7] = 0;
    MEM_OAM[11] = 0;
    MEM_OAM[15] = 128;

    MEM_OAM[19] = 65408;
    MEM_OAM[23] = 0;
    MEM_OAM[27] = 0;
    MEM_OAM[31] = 128;

    MEM_CRAMOBJ[1] = make_color(0,0,20);
    MEM_CRAMOBJ[2] = make_color(0,20,0);
    MEM_CRAMOBJ[3] = make_color(0,20,20);
    MEM_CRAMOBJ[4] = make_color(20,0,0);
    MEM_CRAMOBJ[5] = make_color(20,0,20);
    MEM_CRAMOBJ[6] = make_color(20,15,0);
    MEM_CRAMOBJ[7] = make_color(20,20,20);
    MEM_CRAMOBJ[8] = make_color(0,0,0);
    MEM_CRAMOBJ[9] = make_color(16,16,31);
    MEM_CRAMOBJ[10] = make_color(16,31,16);
    MEM_CRAMOBJ[11] = make_color(16,31,31);
    MEM_CRAMOBJ[12] = make_color(31,16,16);
    MEM_CRAMOBJ[13] = make_color(31,16,31);
    MEM_CRAMOBJ[14] = make_color(31,31,16);
    MEM_CRAMOBJ[15] = make_color(31,31,31);

    MEM_OAM[0] = 120 | 0x4000;
    MEM_OAM[4] = 120 | 0x4000;
    MEM_OAM[8] = 120;
    MEM_OAM[1] = 80 | 0x4000;
    MEM_OAM[5] = 120 | 0x4000;
    MEM_OAM[9] = 152;
    MEM_OAM[2] = 559;
    MEM_OAM[6] = 563;
    MEM_OAM[10] = 564;

    VBlankIntrWait();
    LZ77UnCompVRAM((u32)pmd_titleBitmap, (u32)MEM_VRAM);

    VBlankIntrWait();
    REG_DISPLAYCONTROL = VIDEOMODE_3 | BGMODE_2 | 0x1000;

    keyinput = REG_KEYINPUT;

    gbt_play(template, -1);
    gbt_loop(1);

    for (;;) {

        // Wait until next frame
        VBlankIntrWait();

        // Get buttons state
        oldinput = keyinput;
        keyinput = REG_KEYINPUT;
        newinput = ~((~keyinput) & oldinput);

        if ((newinput & KEY_SELECT) == 0){
            if(gbt_is_playing()){
                gbt_stop();
            }else{
                gbt_play(template, -1);
                gbt_loop(1);
            }
        }
        if ((newinput & KEY_START) == 0)
            break;

    }

    VBlankIntrWait();
    REG_DISPLAYCONTROL = VIDEOMODE_3 | BGMODE_2 | 0x1000 | (1 << 7);

    for (i = 0; i < SCREEN_WIDTH * SCREEN_HEIGHT; ++i)
        MEM_VRAM[i] = make_color(0, 0, 0);

    MEM_OAM[10] = 591;
    MEM_OAM[46] = 519;

    fillRectangle(0, 0, 239, 42, make_color(0,15,28));
    fillRectangle(0, 42, 239, 72, make_color(0,16,29));
    fillRectangle(0, 72, 239, 121, make_color(0,17,30));
    fillRectangle(0, 120, 239, 159, make_color(16,16,16));
    drawRectangle(0, 120, 239, 159, make_color(15,15,16));

    backupCo = MEM_VRAM[x + y * SCREEN_WIDTH];

    VBlankIntrWait();
    REG_DISPLAYCONTROL = VIDEOMODE_3 | BGMODE_2 | 0x1000;

    gameTimer = 0x200;

    // Loop forever
    for (;;) {

        // Wait until next frame
        VBlankIntrWait();
        gameTimerFrames++;

        // Get buttons state
        oldinput = keyinput;
        keyinput = REG_KEYINPUT;
        newinput = ~((~keyinput) & oldinput);

        // Paint old pixel black
        MEM_VRAM[x + y * SCREEN_WIDTH] = backupCo;

        // Handle buttons
        if ((keyinput & KEY_RIGHT) == 0)
            x += 1;
        if ((keyinput & KEY_LEFT) == 0)
            x -= 1;
        if ((keyinput & KEY_DOWN) == 0)
            y += 1;
        if ((keyinput & KEY_UP) == 0)
            y -= 1;
        if ((newinput & KEY_SELECT) == 0){
            if(gbt_is_playing()){
                gbt_stop();
            }else{
                gbt_play(template, -1);
                gbt_loop(1);
            }
        }
        if ((newinput & KEY_START) == 0)
            break;
        if ((((newinput & KEY_A) == 0) || ((newinput & KEY_B) == 0)) && (watergunTimer == 0)){
                watergunTimer = 80;
                if(check_point_collision(tx-16,ty-16,32,32,x,y)){
                        tTimer = 1;
                        if(tt < 127){
                                score++;
                                if((score & 0x0f) > 9){
                                        score += 0x06;
                                }
                        }else{
                                score--;
                                if((score & 0x0f) > 9){
                                        score -= 0x06;
                                }
                                if(score > 0x99){
                                        score = 0;
                                }
                        }
                }
        }

        // Limit possible range
        if (x < 4)
            x = 4;
        if (x >= SCREEN_WIDTH)
            x = SCREEN_WIDTH-1;
        if (y < 4)
            y = 4;
        if (y >= SCREEN_HEIGHT)
            y = SCREEN_HEIGHT-1;

        if(x < 80){
                MEM_OAM[1] = (56) | 0x8000;
                MEM_OAM[2] = 512;
                MEM_OAM[4] = 112 | 0x8300;
                MEM_OAM[5] = 88 | 0x4000;
                MEM_OAM[6] = 514;
        }else if (x > 160){
                MEM_OAM[0] = 112 | 0x8300;
                MEM_OAM[1] = (136) | 0x8200;
                MEM_OAM[2] = 513;
                MEM_OAM[4] = 110 | 0x8300;
                MEM_OAM[5] = 168 | 0x4200;
                MEM_OAM[6] = 512;
        }else{
                MEM_OAM[0] = 112 | 0x8300;
                MEM_OAM[1] = (x-16) | 0x8000;
                MEM_OAM[2] = 515;
                MEM_OAM[4] = 0x200;
                MEM_OAM[5] = 0x200;
        }

        if(watergunTimer > 0){
                MEM_OAM[10] = 592;
                watergunTimer--;
        }else{
                MEM_OAM[8] = y-4;
                MEM_OAM[9] = x-4;
                MEM_OAM[10] = 591;
        }

        if(tTimer-- > 0){
                if(tt < 127){
                        MEM_OAM[44] = (ty-16);
                        MEM_OAM[45] = (tx-16) | 0x8000;
                        MEM_OAM[46] = 519;
                }else{
                        MEM_OAM[44] = (ty-12) | 0x8000;
                        MEM_OAM[45] = (tx-8) | 0x8000;
                        MEM_OAM[46] = 517;
                }
                if(tLoc == 4){
                        tx--;
                }
                if(tLoc == 5){
                        tx++;
                }
                ttn += 2;
        }else{
                tTimer = 200;
                tt = ttn;
                tLoc += 1+(score & 0x03)+(x & 0x01);
                if(tLoc > 7){
                        tLoc -= 7;
                }
                tx = targetPos[(tLoc<<1)];
                ty = targetPos[(tLoc<<1)+1];
        }

        if(gameTimerFrames > 59){
                gameTimerFrames = 0;
                gameTimer--;
                if((gameTimer & 0x0f) > 9){
                        gameTimer -= 6;
                }
                if((gameTimer & 0xff) > 0x59){
                        gameTimer -= 0xa0;
                }
        }
        if(gameTimer == 0x00){
                break;
        }

        MEM_OAM[12] = 8 | 0x4000;
        MEM_OAM[13] = 8;
        MEM_OAM[14] = 538;
        MEM_OAM[16] = 16;
        MEM_OAM[17] = 16;
        MEM_OAM[18] = (score & 0x0f)+527;
        MEM_OAM[20] = 16;
        MEM_OAM[21] = 8;
        MEM_OAM[22] = ((score & 0xf0)>>4)+527;

        MEM_OAM[24] = 16;
        MEM_OAM[25] = 208;
        MEM_OAM[26] = 537;
        MEM_OAM[28] = 16;
        MEM_OAM[29] = 200;
        MEM_OAM[30] = ((gameTimer & 0xf00)>>8)+527;
        MEM_OAM[32] = 16;
        MEM_OAM[33] = 216;
        MEM_OAM[34] = ((gameTimer & 0xf0)>>4)+527;
        MEM_OAM[36] = 16;
        MEM_OAM[37] = 224;
        MEM_OAM[38] = (gameTimer & 0x0f)+527;
        MEM_OAM[40] = 8 | 0x4000;
        MEM_OAM[41] = 200 | 0x4000;
        MEM_OAM[42] = 540;
        

        // Paint pixel white
        backupCo = MEM_VRAM[x + y * SCREEN_WIDTH];
        MEM_VRAM[x + y * SCREEN_WIDTH] = make_color(31, 31, 31);
    }

    VBlankIntrWait();
    REG_DISPLAYCONTROL = VIDEOMODE_3 | BGMODE_2 | (1 << 7);

    VBlankIntrWait();
    LZ77UnCompVRAM((u32)pmd_adBitmap, (u32)MEM_VRAM);

    VBlankIntrWait();
    REG_DISPLAYCONTROL = VIDEOMODE_3 | BGMODE_2;

    for (;;) {

        // Wait until next frame
        VBlankIntrWait();

        // Get buttons state
        oldinput = keyinput;
        keyinput = REG_KEYINPUT;
        newinput = ~((~keyinput) & oldinput);

        if ((newinput & KEY_SELECT) == 0){
            if(gbt_is_playing()){
                gbt_stop();
            }else{
                gbt_play(template, -1);
                gbt_loop(1);
            }
        }
        if ((keyinput & (KEY_START|KEY_A)) == 0){
            __asm(
		"mov r0,#0xF6" "\n"	//Reset all except for VRAM and EWRAM
		"swi 0x01" "\n"		//RegisterRamReset
		"mov r0,#0x01" "\n"	//Reboot at EWRAM
		"ldr r1,=0x03007FFA" "\n"	//Return type flag
		"strb r0,[r1]" "\n"			//Set value
		"swi 0x00" "\n"		//Soft Reset
		".ltorg"
            );
        }

    }

    return 0;
}
